@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2018 Marco d'Itri <md@linux.it>.
+ * Copyright (C) 1999-2022 Marco d'Itri <md@linux.it>.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -61,7 +61,7 @@
 #endif
 
 /* Global variables */
-int sockfd, verb = 0;
+int sockfd, verb = 0, no_recursion = 0;
 
 #ifdef ALWAYS_HIDE_DISCL
 int hide_discl = HIDE_NOT_STARTED;
@@ -84,6 +84,7 @@ int main(int argc, char *argv[])
 	{"version",		no_argument,		NULL, 1  },
 	{"verbose",		no_argument,		NULL, 2  },
 	{"help",		no_argument,		NULL, 3  },
+	{"no-recursion",	no_argument,		NULL, 4  },
 	{"server",		required_argument,	NULL, 'h'},
 	{"host",		required_argument,	NULL, 'h'},
 	{"port",		required_argument,	NULL, 'p'},
@@ -159,7 +160,7 @@ int main(int argc, char *argv[])
     argv = merge_args(getenv("WHOIS_OPTIONS"), argv, &argc);
 
     while ((ch = GETOPT_LONGISH(argc, argv,
-		"abBcdFg:Gh:Hi:KlLmMp:q:rRs:t:T:v:V:x",
+		"abBcdFg:Gh:Hi:IKlLmMp:q:rRs:t:T:v:V:x",
 		longopts, &longindex)) > 0) {
 	/* RIPE flags */
 	if (strchr(ripeflags, ch)) {
@@ -221,9 +222,15 @@ int main(int argc, char *argv[])
 	case 'H':
 	    hide_discl = HIDE_NOT_STARTED;	/* enable disclaimers hiding */
 	    break;
+	case 'I':
+	    server = strdup("\x0E");
+	    break;
 	case 'p':
 	    port = strdup(optarg);
 	    break;
+	case 4:
+	    no_recursion = 1;
+	    break;
 	case 3:
 	    usage(EXIT_SUCCESS);
 	case 2:
@@ -338,13 +345,17 @@ int handle_query(const char *hserver, const char *hport,
 	    sockfd = openconn(server + 1, NULL);
 	    free(server);
 	    server = query_crsnic(sockfd, query);
+	    if (no_recursion)
+		server[0] = '\0';
 	    break;
 	case 8:
 	    if (verb)
-		printf(_("Using server %s.\n"), "whois.afilias-grs.info");
-	    sockfd = openconn("whois.afilias-grs.info", NULL);
+		printf(_("Using server %s.\n"), server + 1);
+	    sockfd = openconn(server + 1, NULL);
 	    free(server);
 	    server = query_afilias(sockfd, query);
+	    if (no_recursion)
+		server[0] = '\0';
 	    break;
 	case 0x0A:
 	    p = convert_6to4(query);
@@ -372,6 +383,13 @@ int handle_query(const char *hserver, const char *hport,
 	    server = guess_server(p);
 	    free(p);
 	    goto retry;
+	case 0x0E:
+	    if (verb)
+		printf(_("Using server %s.\n"), "whois.iana.org");
+	    sockfd = openconn("whois.iana.org", NULL);
+	    free(server);
+	    server = query_iana(sockfd, query);
+	    break;
 	default:
 	    break;
     }
@@ -394,7 +412,7 @@ int handle_query(const char *hserver, const char *hport,
     free(query_string);
 
     /* recursion is fun */
-    if (server && !strchr(query, ' ')) {
+    if (!no_recursion && server && !strchr(query, ' ')) {
 	printf(_("\n\nFound a referral to %s.\n\n"), server);
 	handle_query(server, NULL, query, flags);
     }
@@ -475,7 +493,7 @@ const char *match_config_file(const char *s)
 #endif
 
 /* Parses an user-supplied string and tries to guess the right whois server.
- * Returns a dinamically allocated buffer.
+ * Returns a dynamically allocated buffer.
  */
 char *guess_server(const char *s)
 {
@@ -683,7 +701,10 @@ char *queryformat(const char *server, const char *flags, const char *query)
 	    strcat(buf, "a ");
 	    strcat(buf, query + 2);
 	} else if (myinet_aton(query) || strchr(query, ':')) {
-	    strcat(buf, "n + ");
+	    if (strchr(query, '/'))
+		strcat(buf, "r + = ");
+	    else
+		strcat(buf, "n + ");
 	    strcat(buf, query);
 	} else
 	    strcat(buf, query);
@@ -921,6 +942,47 @@ char *query_afilias(const int sock, const char *query)
     return referral_server;
 }
 
+char *query_iana(const int sock, const char *query)
+{
+    char *temp, *p, buf[2000];
+    FILE *fi;
+    char *referral_server = NULL;
+    int state = 0;
+
+    temp = malloc(strlen(query) + 2 + 1);
+    strcpy(temp, query);
+    strcat(temp, "\r\n");
+
+    fi = fdopen(sock, "r");
+    if (write(sock, temp, strlen(temp)) < 0)
+	err_sys("write");
+    free(temp);
+
+    while (fgets(buf, sizeof(buf), fi)) {
+	/* If multiple attributes are returned then use the first result.
+	   This is not supposed to happen. */
+	if (state == 0 && strneq(buf, "refer:", 6)) {
+	    for (p = buf; *p != ':'; p++);	/* skip until colon */
+	    for (p++; *p == ' '; p++);		/* skip colon and spaces */
+	    referral_server = strdup(p);
+	    if ((p = strpbrk(referral_server, "\r\n ")))
+		*p = '\0';
+	    state = 2;
+	}
+
+	if ((p = strpbrk(buf, "\r\n")))
+	    *p = '\0';
+	recode_fputs(buf, stdout);
+	fputc('\n', stdout);
+    }
+
+    if (ferror(fi))
+	err_sys("fgets");
+    fclose(fi);
+
+    return referral_server;
+}
+
 int openconn(const char *server, const char *port)
 {
     int fd = -1;
@@ -1073,7 +1135,7 @@ int japanese_locale(void)
 {
     char *lang;
 
-    lang = getenv("LC_MESSAGE");
+    lang = getenv("LC_MESSAGES");
     if (lang) {
 	if (strneq(lang, "ja", 2))
 	    return 0;
@@ -1437,31 +1499,47 @@ void NORETURN usage(int error)
 "Usage: whois [OPTION]... OBJECT...\n\n"
 "-h HOST, --host HOST   connect to server HOST\n"
 "-p PORT, --port PORT   connect to PORT\n"
+"-I                     query whois.iana.org and follow its referral\n"
 "-H                     hide legal disclaimers\n"
+    ));
+    fprintf((EXIT_SUCCESS == error) ? stdout : stderr, _(
 "      --verbose        explain what is being done\n"
+"      --no-recursion   disable recursion from registry to registrar servers\n"
 "      --help           display this help and exit\n"
 "      --version        output version information and exit\n"
 "\n"
+    ));
+    fprintf((EXIT_SUCCESS == error) ? stdout : stderr, _(
 "These flags are supported by whois.ripe.net and some RIPE-like servers:\n"
 "-l                     find the one level less specific match\n"
 "-L                     find all levels less specific matches\n"
 "-m                     find all one level more specific matches\n"
 "-M                     find all levels of more specific matches\n"
+    ));
+    fprintf((EXIT_SUCCESS == error) ? stdout : stderr, _(
 "-c                     find the smallest match containing a mnt-irt attribute\n"
 "-x                     exact match\n"
 "-b                     return brief IP address ranges with abuse contact\n"
+    ));
+    fprintf((EXIT_SUCCESS == error) ? stdout : stderr, _(
 "-B                     turn off object filtering (show email addresses)\n"
 "-G                     turn off grouping of associated objects\n"
 "-d                     return DNS reverse delegation objects too\n"
+    ));
+    fprintf((EXIT_SUCCESS == error) ? stdout : stderr, _(
 "-i ATTR[,ATTR]...      do an inverse look-up for specified ATTRibutes\n"
 "-T TYPE[,TYPE]...      only look for objects of TYPE\n"
 "-K                     only primary keys are returned\n"
 "-r                     turn off recursive look-ups for contact information\n"
+    ));
+    fprintf((EXIT_SUCCESS == error) ? stdout : stderr, _(
 "-R                     force to show local copy of the domain object even\n"
 "                       if it contains referral\n"
 "-a                     also search all the mirrored databases\n"
 "-s SOURCE[,SOURCE]...  search the database mirrored from SOURCE\n"
 "-g SOURCE:FIRST-LAST   find updates from SOURCE from serial FIRST to LAST\n"
+    ));
+    fprintf((EXIT_SUCCESS == error) ? stdout : stderr, _(
 "-t TYPE                request template for object of TYPE\n"
 "-v TYPE                request verbose template for object of TYPE\n"
 "-q [version|sources|types]  query specified server info\n"
